use std::sync::Arc;

use bevy_ecs::world::World;
use engine_window::raw_handle::DisplayHandleWrapper;
use vulkano::{
    VulkanLibrary,
    instance::{Instance, InstanceCreateFlags, InstanceCreateInfo, InstanceExtensions},
    swapchain::Surface,
};

use crate::{VulkanConfig, VulkanInstance};

fn load_library() -> Arc<VulkanLibrary> {
    let library = VulkanLibrary::new().unwrap();

    log::debug!("Available Instance layers:");
    for layer in library.layer_properties().unwrap() {
        log::debug!(
            "\t - Layer name: {}, Description: {}, Implementation Version: {}, Vulkan Version: {}",
            layer.name(),
            layer.description(),
            layer.implementation_version(),
            layer.vulkan_version()
        );
    }

    library
}

pub fn create_and_insert_instance(world: &mut World, config: &VulkanConfig) {
    let library = load_library();

    let instance_extensions = {
        if config.with_window_surface {
            let display_handle = world
                .get_resource::<DisplayHandleWrapper>()
                .expect("DisplayHandleWrapper must be added before VulkanPlugin");

            Surface::required_extensions(&display_handle.0)
                .expect("Failed to get surface required extensions")
        } else {
            InstanceExtensions::default()
        }
    };

    let instance = Instance::new(
        library,
        InstanceCreateInfo {
            // Enable enumerating devices that use non-conformant Vulkan implementations.
            // (e.g. MoltenVK)
            flags: InstanceCreateFlags::ENUMERATE_PORTABILITY,
            enabled_extensions: instance_extensions,
            enabled_layers: config.instance_layers.clone(),
            ..Default::default()
        },
    )
    .expect("Failed to create vulkan instance");

    log::debug!("Instance created");
    log::debug!(
        "\t- Enabled extensions: {:?}",
        instance.enabled_extensions()
    );
    log::debug!("\t- Enabled layers: {:?}", instance.enabled_layers());
    log::debug!("\t- API version: {:?}", instance.api_version());
    log::debug!("\t- Max API version: {:?}", instance.max_api_version());
    log::debug!("\t- Flags: {:?}", instance.flags());

    world.insert_resource(VulkanInstance(instance));
}
